module NPGRT
	module Algorithm
		def dfs(start, &b)
			yield(lambda{|v| dfs(v, &b)}, start)
		end

		def bfs(*start)
			yield(start.method(:push), start.shift)	until start.empty?
		end

		def array_to_hash(arr)
			hsh = {}
			arr.each_with_index{|e, i| hsh[i] = e if e}
			hsh	
		end

		def hash_to_array(hsh)
			arr = []
			hsh.each{|k, v| arr[k] = v}
			arr
		end

		class AlgorithmError < StandardError; end
		class CycleFound < StandardError; end
		class NotFound < StandardError; end
		def graph_buildpath(st, ed, pre)
			ret = []
			while ed != st && ed != -1
				raise CycleFound, 'path cycled' if ret.include?(ed)
				ret.unshift ed
				ed = pre[ed]
			end	
			if ed == st
				ret.unshift st
			else
				raise NotFound
			end	
			ret		
		end


		class Graph
			attr_accessor :adjlist, :adjmat

			RelaxOperators = Struct.new(:add, :min, :maxvalue)
			DefaultRelax = RelaxOperators.new(
				lambda{|a, b| a + b},
				lambda{|a, b| [a,b].min},
				Float::INFINITY
			)

			def adjlist
				@adjlist ||= adjmat.map(&NPGRT.method(:array_to_hash))
			end
			def adjmat
				@adjmat ||= adjlist.map(&NPGRT.method(:hash_to_array))
			end
			def floyd!(relax = DefaultRelax)
				a = adjmat
				n = adjmat.size
				self.adjlist = nil #reset
				(0...n).each{|k|
					(0...n).each{|i|
						next if i == k
						next if a[i][k] == nil
						(0...n).each{|j|
							next if j == k || j == i
						        next if a[i][k] == nil || a[k][j] == nil
							if !a[i][j]
								a[i][j] = relax.add[ a[i][k], a[k][j] ]
							else	
								a[i][j] = relax.min[
									a[i][j],
							       		relax.add[a[i][k], a[k][j] ] 

								]
							end
						}
					}
				}
				self
			end

			def floyd
				deepcopy_adjmat.floyd!
			end

			def bellman_ford(start = [0], relax = DefaultRelax)
				a   = self.adjlist
				dis = a.map { relax.maxvalue }
				pre = a.map { -1 }
				start.each{|i| dis[i] = 0; pre[i] = i }
				n   = a.size
				(n-1).times{
					relaxed = false
					each_edge do |u, v, w|
							next if w == nil
							rel = relax.add[ dis[u] , w]
							if rel == relax.min[ rel, dis[v] ]
								dis[v] = rel
								pre[v] = u
								relaxed = true
							end
					end	
					break if !relaxed
				}
				[dis, pre]
			end

			def spfa(start = [0], relax = DefaultRelax)
				a   = self.adjlist
				dis = a.map { relax.maxvalue }
				pre = a.map { -1 }
				inq = []
				start.each{|i| dis[i] = 0; pre[i] = i; inq[i] = true }
				n   = a.size
				
				NPGRT.bfs *start do |yielder, u|
					inq[u] = false
					a[u].each do |v, w|
						next if w == nil
						rel = relax.add[ dis[u] , w]
						if rel == relax.min[ rel, dis[v] ]
							dis[v] = rel
							pre[v] = u
							if !inq[v]
								inq[v] = true
								yielder.call v
							end
						end 
					end
				end

				[dis, pre]
			end

			def each_edge(start = (0...adjlist.size))
				start.each{|u| self.adjlist[u].each{|v, w|
					yield u, v, w
				}}
			end

			def toposort
				x = self.adjlist
				id = x.map { 0 }
			 	each_edge{|u, v, w| id[v] += 1 if w}
				n = x.size
				ret = []
				while true
				   a = (0...n).select{|u| id[u] == 0} - ret
				   break if a == []
				   ret.concat a
				   each_edge a do |u, v, w|
 					id[v] -= 1 if w
				   end
				end
				raise CycleFound, 'in toposorting' if ret.size != n
				ret
			end
			
			def deepcopy_adjmat
				x = Graph.new
				x.adjmat = Marshal.load Marshal.dump adjmat
				x.adjlist = nil
				x
			end

			def deepcopy_adjlist
				x = Graph.new
				x.adjlist = Marshal.load Marshal.dump adjlist
				x.adjmat  = nil
				x
			end
		end
	end
end


